(*****************************************************************************
 *          HOL-OCL - a shallow embedding of OCL in Isabelle/HOL              
 *                                                                            
 * Lifting.thy - basic type construction via strictification, smashing and 
 *               lifting.
 * Copyright (C) 2001-2006 Achim D. Brucker <brucker@inf.ethz.ch>   
 *                         Burkhart Wolff   <bwolff@inf.ethz.ch>    
 *                                                                            
 * This file is part of HOL-OCL.                                              
 *                                                                            
 * HOL-OCL is free software; you can redistribute it and/or modify it under   
 * the terms of the GNU General Public License as published by the Free       
 * Software Foundation; either version 2 of the License, or (at your option)  
 * any later version.                                                         
 *                                                                            
 * HOL-OCL is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS 
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more 
 * details.                                                              
 *                                                                            
 * You should have received a copy of the GNU General Public License along    
 * with this program; if not, write to the Free Software Foundation, Inc.,    
 * 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.                  
 ******************************************************************************)

header {* The Theory of Lifting and its Combinators *} 

theory Lifting 
imports 
  Main 
begin 
text_raw{* \label{thy:OCL_Lifting} *}
text {*
  Besides the infrastructure for handling
  undefinedness\index{undefined}, this theory also provides the basis
  for data type embedding smashing\index{smashing} and context
  lifting\index{lifting} including the generic core of theorems.

  The main purpose of this theory is to provide a generic theory of
  undefinedness. The Isabelle standard mechanism for such a generic
  data type is the class mechanism. The standard method is to declare
  a class to which all related operations are associated.
*}



section {* A Generic Theory of Undefinedness *} 
text {* 
  Since the very first \OCL{} 
  publictions~\cite{omg:uml:1999} explicit
  undefinedness\index{undefinedness} is part of the language, both for
  the logic and the basic values, \eg,~\cite[7.4.9 Undefined
  Values]{omg:uml:1999} states (this is similary stated
  in~\cite[page 15]{omg:ocl:2003}):
  \begin{oclspecification}[omg:uml:1999]{7-11} 
    Whenever an \OCL expressionis beeing evaluated, there is a possibility
    that one or more of the queries in the expression are undefined. If
    this is the case, then the complete expression will be undefined.

    There are two exceptions to this for the Boolean operators
    \begin{itemize}
      \item True \textsc{or}-ed with anything is True
      \item False \textsc{and}-ed with anything is False
    \end{itemize}
    The above two rules are valid irrespective of the order of the 
    arguments and the above rules are valid wether or not the value 
    of the other sub-expression is known.
  \end{oclspecification}
  This requirement postulates the strictness\index{strictness} of all 
  operations (except for the logic) and rules out a modeling of 
  undefinedness via Hilbert-Operators and underspecification (e.g., as 
  done for Z~\cite{iso:z:2002}in 
  Isabelle/\holz~\cite{brucker.ea:hol-z:2003}).
*}

classes   bot0  \<subseteq>  ord          

text {* All bottom types posses a constant $\bottom$ *}
consts    UU :: "'\<alpha>\<Colon>bot0" 
syntax    UU :: "'\<alpha>\<Colon>bot0"   ("\<bottom>")

axclass   bot \<subseteq> bot0
  nonEmpty : "\<exists> x. x \<noteq> \<bottom>"


constdefs 
  DEF            :: "'\<alpha>\<Colon>bot  \<Rightarrow> bool"      
  "DEF x         \<equiv> (x \<noteq> \<bottom> )"              

  isStrict       :: "('\<alpha>\<Colon>bot \<Rightarrow> '\<beta>\<Colon>bot) \<Rightarrow> bool" 
  "isStrict f    \<equiv> (f \<bottom> = \<bottom> )"   

  strictify      :: "('\<alpha>::bot \<Rightarrow> '\<beta>\<Colon>bot) \<Rightarrow> '\<alpha> \<Rightarrow> '\<beta>"
  "strictify f x \<equiv> if x=\<bottom> then \<bottom> else f x"

  smash          :: "[['\<beta>::bot, '\<alpha>\<Colon>bot] \<Rightarrow> bool, '\<alpha>] \<Rightarrow> '\<alpha>" 
  "smash f X     \<equiv> if f \<bottom> X then \<bottom> else X"

section {* Semantic Constructions: Liftings, Functions, Products *}


text {* 
  We introduce now the lifting construction 
  (see Winskel~\cite[p.131]{winskel:semantics:1993}) by a type constructor 
  defined as free data type:
*}

datatype   '\<alpha> up = lift '\<alpha> | down

constdefs  
   drop :: "'\<alpha> up \<Rightarrow> '\<alpha>"
  "drop x  \<equiv> case x of lift v \<Rightarrow> v | down \<Rightarrow> \<some>  x . True"

syntax     
  "@lift"        :: "'\<alpha> \<Rightarrow> '\<alpha> up"   ("|.(_).|") 
  "Lifting.drop" :: "'\<alpha> up \<Rightarrow> '\<alpha>"   ("|^(_)^|")

 
syntax (xsymbols)
  "@lift"         :: "'\<alpha> \<Rightarrow> '\<alpha> up"   ("\<HolOclLiftLeft>_\<HolOclLiftRight>")  
  "Lifting.drop"  :: "'\<alpha> up \<Rightarrow> '\<alpha>"   ("\<HolOclDropLeft>_\<HolOclDropRight>")

translations
  "|.a.|" \<rightleftharpoons> "lift a"
  "|^a^|" \<rightleftharpoons> "drop a"


instance   up   :: (zero) zero
  by intro_classes
instance   up   :: (one) one
  by intro_classes
instance   up   :: (plus) plus
  by intro_classes
instance   up   :: (minus) minus
  by intro_classes
instance   up   :: (times) times
  by intro_classes
instance   up   :: (inverse) inverse
  by intro_classes

section {* Semantic Constructions: Liftings of Type Constructors *}
text {* The class is then propagated across lifting, function space
and cartesian products. *}
instance   up   :: (type) ord
  by intro_classes
arities    up   :: (type) bot         
instance   fun  :: (type,ord) ord     
  by intro_classes
arities    fun  :: (type,bot) bot     
instance   "*"  :: (ord,ord)  ord
  by intro_classes
instance   "+"  :: (ord,ord)  ord
  by intro_classes

arities    "*"  :: (bot,bot)  bot     
arities    "+"  :: (bot,bot)  bot     

defs       
  UU_up_def[simp]:   "\<bottom> \<equiv> down"
  UU_fun_def[simp]:  "\<bottom> \<equiv> (\<lambda>x. \<bottom>)"
  UU_pair_def[simp]: "\<bottom> \<equiv> (\<bottom>, \<bottom>)"
  UU_sum_def[simp]:  "\<bottom> \<equiv> (Inl \<bottom>)"

instance   up   :: (type) bot        
  by intro_classes
instance   fun  :: (type,bot) bot     
  by intro_classes
instance   "*"  :: (bot,bot)  bot    
  by intro_classes
instance   "+"   :: (bot,bot) bot       
  by intro_classes

section {* Semantic Constructions Dealing with Context Lifting  *}

text {* \todo{Introduce syntax translation for VAL} *}
types     ('\<tau>, 'b) VAL = "'\<tau> \<Rightarrow> 'b" 
 
instance fun::(type,zero) zero 
  by intro_classes
instance fun::(type,one)  one  
  by intro_classes
instance   fun:: (type,plus) plus
  by intro_classes
instance   fun:: (type,minus) minus
  by intro_classes
instance   fun:: (type,times) times
  by intro_classes
instance   fun:: (type,inverse) inverse
  by intro_classes
instance   fun:: (type,ord) ord
  by intro_classes


constdefs  
  lift0  :: "'\<alpha> \<Rightarrow> ('\<tau>,'\<alpha>) VAL"
  "lift0 \<equiv> \<lambda>c. \<lambda>s. c"
  lift1  :: "('\<alpha> \<Rightarrow> '\<beta>) \<Rightarrow> ('\<tau>, '\<alpha>) VAL  \<Rightarrow> ('\<tau>, '\<beta>) VAL"
  "lift1 \<equiv> (\<lambda>f X \<tau> . f (X \<tau>))"
  lift2  :: "(['\<alpha>,'\<beta>]\<Rightarrow>'\<gamma>) \<Rightarrow> [('\<tau>,'\<alpha>)VAL,('\<tau>,'\<beta>)VAL]\<Rightarrow>('\<tau>,'\<gamma>)VAL"
  "lift2 \<equiv> (\<lambda>f X Y \<tau>. f (X \<tau>)(Y \<tau>))"
  lift3  :: "(['\<alpha>,'\<beta>, '\<gamma>]\<Rightarrow>'\<delta>) \<Rightarrow> [('\<tau>,'\<alpha>)VAL,('\<tau>,'\<beta>)VAL, ('\<tau>,'\<gamma>)VAL]\<Rightarrow>('\<tau>,'\<delta>)VAL"
  "lift3 \<equiv> (\<lambda>f X Y Z \<tau>. f (X \<tau>)(Y \<tau>)(Z \<tau>))"

  cp     :: "([('\<tau>,'\<alpha>) VAL, '\<tau>] \<Rightarrow> '\<beta>) \<Rightarrow> bool"
  "cp(P) \<equiv> (\<exists> f. \<forall> X \<tau>. P X \<tau> = f (X \<tau>) \<tau>)" 
 

syntax      
  lift0  :: "'\<alpha> \<Rightarrow> ('\<tau>,'\<alpha>) VAL" 
            ("lift\<^sub>0")
  lift1  :: "('\<alpha> \<Rightarrow> '\<beta>) \<Rightarrow> ('s, '\<alpha>) VAL  \<Rightarrow> ('s, '\<beta>) VAL" 
            ("lift\<^sub>1")
  lift2  :: "(['\<alpha>,'\<beta>]\<Rightarrow>'\<gamma>) \<Rightarrow> [('\<tau>,'\<alpha>)VAL,('\<tau>,'\<beta>)VAL]\<Rightarrow>('\<tau>,'\<gamma>)VAL" 
            ("lift\<^sub>2")
  lift3  :: "(['\<alpha>,'\<beta>, '\<gamma>]\<Rightarrow>'\<delta>) \<Rightarrow> [('\<tau>,'\<alpha>)VAL,('\<tau>,'\<beta>)VAL, ('\<tau>,'\<gamma>)VAL]\<Rightarrow>('\<tau>,'\<delta>)VAL" 
            ("lift\<^sub>3")

section {* A Generic Theory of Undefinedness, Strictness, and Smashing  *}

lemma not_DEF_UU [simp]: "\<not> DEF(\<bottom>)"
  by (simp (no_asm) add: DEF_def)

lemma not_DEF_X: "(\<not> DEF X) = (X = \<bottom>)";
  by (simp (no_asm) add: DEF_def)

lemma exists_DEF: "\<exists> x. DEF(x)"
  by (simp add: DEF_def nonEmpty)


lemma isStrict_strictify [simp]: "isStrict(strictify f)"
  by (simp add: isStrict_def strictify_def)


lemma strict2a_UU [simp]: "strictify f \<bottom> = \<bottom>"
  by (simp (no_asm) add: strictify_def)

lemma strict2b_UU [simp]: "strictify f \<bottom> X = \<bottom>"
  by (simp (no_asm) add: strictify_def UU_fun_def)



section {* Strictness versus Definedness  *}

lemma DEF_strictify_DEF_args2 : 
  "DEF (strictify (\<lambda>x. strictify (f x)) X Y) \<Longrightarrow> DEF X \<and> DEF Y"
  apply (simp add: strictify_def DEF_def)
  apply (case_tac "X=\<bottom>")
  apply auto
  done


lemma DEF_strictify_DEF_fun : 
  "\<lbrakk> \<And>X. DEF(f X) ; DEF X \<rbrakk> \<Longrightarrow> DEF(strictify f X)"
  by (simp add: strictify_def DEF_def )


lemma DEF_strictify_DEF_args : 
  "DEF(strictify f X) \<Longrightarrow>  DEF f \<and> DEF X"
  by (simp add: strictify_def DEF_def, auto)


lemma isStrict_compose : 
  "\<lbrakk> isStrict f; isStrict g \<rbrakk> \<Longrightarrow> isStrict (f o g)"
  by (simp add: isStrict_def o_def)

lemma smash_strict [simp]:  "smash f \<bottom> = \<bottom>"
  by (simp (no_asm) add: smash_def)

lemma smashed_sets_nonempty [simp]: "\<bottom> : {X. smash f X = X}"
  by (simp (no_asm))

text {* 
  This lemmas are useful for proofs of non-emptyness of type-definition 
  based on smashed collection types.
*}



section {* A Theory of Undefinedness and Strictness in Lifted Types *}

lemma not_down_exists_lift :  "x \<noteq>  down = (\<exists>y. x = \<HolOclLiftLeft>y\<HolOclLiftRight>)"
  by (induct_tac "x",auto)

lemma not_down_exists_lift2 : "x \<noteq>  \<bottom> = (\<exists> y. x = \<HolOclLiftLeft>y\<HolOclLiftRight>)"
  by (induct_tac "x", auto)

lemma drop_lift [simp]: "\<HolOclDropLeft>\<HolOclLiftLeft>f\<HolOclLiftRight>\<HolOclDropRight> = f"
  by (simp add: drop_def UU_up_def)

lemma drop_down [simp]: "\<HolOclDropLeft>\<bottom>\<HolOclDropRight> = (\<some> x. True)"
  by (simp add: drop_def)

lemma drop_down2 [simp]: "\<HolOclDropLeft>down\<HolOclDropRight> = (\<some> x. True)"
  by (simp add: drop_def)


lemma not_DEF_down [simp]: "\<not>DEF(down)"
  by (simp (no_asm) add: DEF_def)

lemma DEF_lift [simp]: "DEF(\<HolOclLiftLeft>x\<HolOclLiftRight>)"
  by (simp (no_asm) add: DEF_def)


lemma DEF_X_up : "DEF(X::'\<alpha> up) = (\<exists> x. X = \<HolOclLiftLeft>x\<HolOclLiftRight>)"
  by (simp add: not_down_exists_lift DEF_def) 


lemma not_DEF_X_up : "(\<not>DEF(X\<Colon>'\<alpha> up)) = (X = \<bottom>)"
  by (simp (no_asm) add: DEF_def)


lemma DEF_fun_lift [simp]: "DEF(\<lambda>x. \<HolOclLiftLeft>f x\<HolOclLiftRight>)"
  apply (simp (no_asm) add: DEF_def)
  apply (rule notI)
  apply (drule fun_cong)
  apply simp
  done

lemma DEF_fun_fun_lift [simp]: "DEF(\<lambda>x y. \<HolOclLiftLeft>f x y\<HolOclLiftRight>)"
  apply (simp (no_asm) add: DEF_def)
  apply (rule notI)
  apply (drule fun_cong)
  apply (drule fun_cong)
  apply simp
  done

lemma lift_defined : " \<bottom> \<notin> X \<longrightarrow> \<HolOclLiftLeft>X\<HolOclLiftRight> \<noteq> \<bottom>" 
  by(auto)



section {* A Generic Theory of Strictness  *}

lemma strict2c_UU [simp]:  "strictify f down = down"
  by (simp (no_asm) add: strictify_def)

lemma strict2d_UU [simp]: "strictify f down X = down"
  by (simp (no_asm) add: strictify_def)


lemma strict2e_UU [simp]: "strictify (\<lambda>x. strictify (f x)) (X\<Colon>'\<alpha>\<Colon>bot) \<bottom> = \<bottom>"
  by (simp add: strictify_def )

lemma strict2f_UU [simp]: "strictify (\<lambda>x. strictify (f x)) (X\<Colon>'\<alpha>\<Colon>bot) down = down"
  by (simp add: strictify_def )


lemma strict2_DEF [simp]: "DEF X \<Longrightarrow> strictify f X = f X"
  by (simp add: strictify_def, auto)

lemma strict3_DEF [simp]: 
  "\<lbrakk> DEF X; DEF Y \<rbrakk> \<Longrightarrow> strictify(\<lambda> X. strictify (f X)) X Y = f X Y"
  by (simp add: strictify_def, auto)



section {* Generic Undefinedness-Reduction Rules for 2-Lifted Strict Operations *}

lemma lift1b_undef [simp]: "lift\<^sub>1 (strictify f) \<bottom> = \<bottom>"
  by (simp add: lift1_def)

lemma lift2_undef1a [simp]: "lift\<^sub>2 (strictify(\<lambda>x. strictify(f x))) \<bottom> (X\<Colon>('a\<Rightarrow>('b\<Colon>bot))) = \<bottom>"
  by (simp add: lift1_def lift2_def)

lemma lift2_undef2a [simp]: "lift\<^sub>2 (strictify(\<lambda>x. strictify(f x))) (X\<Colon>('\<tau>\<Rightarrow>('b\<Colon>bot))) \<bottom> = \<bottom>"
  by (simp add: lift1_def lift2_def)

lemma lift1_undef_fw: "g \<equiv> lift\<^sub>1 (strictify f) \<Longrightarrow> g \<bottom> = \<bottom>"
  by (simp del: UU_fun_def)

lemma lift2_undef1_fw: "g \<equiv> lift\<^sub>2 (strictify(\<lambda>x. strictify(f x))) \<Longrightarrow> g \<bottom> X = \<bottom>"
  by (simp del: UU_fun_def)

lemma lift2_undef2_fw: "g \<equiv> lift\<^sub>2 (strictify(\<lambda>x. strictify(f x))) \<Longrightarrow> g X \<bottom> = \<bottom>"
  by (simp del: UU_fun_def)


section {* Generic Theorems on Context Lifted Combinators *}

lemma cp_charn : "\<lbrakk> A \<tau> = B \<tau>; cp P \<rbrakk> \<Longrightarrow> P A \<tau> = P B \<tau>";
  by (auto simp: cp_def) 

lemma cp_by_cpify: "cp P = (\<forall> X \<tau>. P X \<tau>  = P (lift0 (X \<tau>)) \<tau>)"
  by(auto simp: cp_def lift0_def, rule exI, assumption)

lemma cp_lift0 [simp, intro!]:  "cp(lift\<^sub>0 c)"
  by (simp add: cp_def lift0_def, fast)

lemma cp_eq [simp, intro!]:
  "\<lbrakk> cp P; cp P'\<rbrakk> \<Longrightarrow> cp (\<lambda>u ua. P u ua = P' u ua)"
  by(auto simp: cp_def)

lemma cp_const [simp, intro!]: "cp(\<lambda>X. c)"
  by (simp add: cp_def, fast)

lemma cp_id [simp, intro!]:  "cp(\<lambda>X. X)"
  by (simp add: cp_def, fast)

lemma cp_compose [simp,intro!]: 
  "\<lbrakk> cp P; cp P' \<rbrakk> \<Longrightarrow> cp(P o P')"
  apply (simp add: cp_def o_def)
  apply (erule exE) 
  apply (erule exE)
  apply (simp (no_asm_simp))
  apply fast
  done

lemma cp_compose2: 
  "\<lbrakk>cp P; cp P'\<rbrakk> \<Longrightarrow> cp (\<lambda>x. P (P' x))"
  by (simp add: cp_def o_def, auto)


lemma cp_lift1 [simp]: 
  "cp P \<Longrightarrow> cp (\<lambda>X. lift\<^sub>1 f (P X))"
  apply (simp add: cp_def lift1_def)
  apply (erule exE)
  apply (simp (no_asm_simp))
  apply fast
  done

lemma cp_lift2 [simp]:
  "\<lbrakk>cp P; cp P'\<rbrakk> \<Longrightarrow> cp (\<lambda>X. lift\<^sub>2 f (P X) (P' X))"
  apply (simp add: cp_def lift2_def)
  apply (erule exE) 
  apply (erule exE)
  apply (simp (no_asm_simp))
  apply fast
  done

lemma cp_lift3 [simp]: 
  "\<lbrakk>cp P; cp P'; cp P''\<rbrakk> \<Longrightarrow> cp (\<lambda>X. lift3 f (P X) (P' X) (P'' X))"
  apply (simp add: cp_def lift3_def)
  apply (erule exE)+ 
  apply (simp (no_asm_simp))
  apply fast
  done

lemma cp_lift0_fw : "f \<equiv> lift\<^sub>0 g \<Longrightarrow> cp f"
  by (simp add: lift0_def cp_def)
 
lemma cp_lift1_fw [simp]:
  "\<lbrakk>f \<equiv> lift\<^sub>1 g; cp P\<rbrakk> \<Longrightarrow> cp (\<lambda>X. f (P X))"
  by (simp add: cp_def lift1_def, auto)

lemma cp_lift2_fw : 
  "\<lbrakk>f \<equiv> lift\<^sub>2 g; cp P; cp P'\<rbrakk>
  \<Longrightarrow> cp (\<lambda>X. f (P X) (P' X))"
  by (simp add: cp_def lift2_def, auto)


lemma cp_lift3_fw:
  "\<lbrakk>f \<equiv> lift3 g; cp P; cp P'; cp P''\<rbrakk>
  \<Longrightarrow> cp (\<lambda>X. f (P X) (P' X) (P'' X))"
  by (simp add: lift3_def cp_def,auto)



lemma ocl_undef_split : "X \<tau> = \<bottom> \<or> (\<exists>a. X \<tau> = (\<HolOclLiftLeft>a\<HolOclLiftRight>))"
  by (simp add: UU_up_def,simp add: not_down_exists_lift [symmetric])


lemma ocl_cp_undef_split:
  "\<lbrakk>cp P; cp P'; P \<bottom> = P' \<bottom>; X \<noteq> \<bottom> \<Longrightarrow> P X = P' X\<rbrakk>
  \<Longrightarrow> P X = P' X"
  by (simp add: UU_fun_def, auto)


lemma ocl_cp_subst : 
  "\<lbrakk>X \<tau> = X' \<tau>; P X \<tau> = C; cp P\<rbrakk> \<Longrightarrow> P X' \<tau> = C"
  by (simp add: cp_def, auto)


lemma ocl_cp_subst2 : 
  "\<lbrakk>X \<tau> = X' \<tau>; P X' \<tau> = P' X' \<tau>; cp P; cp P'\<rbrakk> \<Longrightarrow> P X \<tau> = P' X' \<tau>"
  by (simp add: cp_def, auto)


lemma ocl_cp_subst3 : 
  "\<lbrakk>X \<tau> = X' \<tau>; P X' \<tau> = P' X' \<tau>; cp P; cp P'\<rbrakk> \<Longrightarrow> P X \<tau> = P' X \<tau>"
  by (simp add: cp_def, auto)


lemma drop_lift_idem [simp]:"( \<HolOclDropLeft>\<HolOclLiftLeft> x \<HolOclLiftRight>\<HolOclDropRight>) = x"
  by (simp_all split add: up.split)

lemma lift_drop_idem [simp]: "\<lbrakk>DEF(x)\<rbrakk> \<Longrightarrow> (\<HolOclLiftLeft>\<HolOclDropLeft>x\<HolOclDropRight>\<HolOclLiftRight>) = x"
  apply (simp_all add: drop_def DEF_def not_down_exists_lift)
  apply (erule exE, auto)
  done

lemma DEF_set_charn : "(\<forall>x\<in>X. DEF x) = (\<bottom> \<notin> X)"
  by (simp add: DEF_def, auto)


lemma cp_option_case:
"\<lbrakk>cp P; cp P'; cp (\<lambda> X \<tau>.(\<lambda>x. P'' x X \<tau>))\<rbrakk> 
  \<Longrightarrow> cp (\<lambda>X \<tau> . case (P X \<tau> ) of None \<Rightarrow> P' X \<tau>  | Some x  \<Rightarrow> P'' x X \<tau>)"
apply(simp add: cp_def)
apply(auto)

apply(rule_tac x="\<lambda> C \<tau>. case (f C \<tau>) of None \<Rightarrow> (fa C \<tau>) | Some x \<Rightarrow> (fb  C \<tau>) x" in exI)
apply(auto)
apply(rule_tac t="P' X \<tau>" in subst)
apply(rotate_tac 1)
apply(erule allE)
apply(rotate_tac -1)
apply(erule allE)
apply(rule sym)
apply(assumption)

apply(rule_tac t="\<lambda> x. P'' x X \<tau>" in subst)
apply(rotate_tac 2)
apply(erule allE)
apply(rotate_tac -1)
apply(erule allE)
apply(rule sym)
apply(assumption)

apply(rule refl)
done

section {* Higher-order Context-passingness *}
text {* 
   This theory is inteded to develop higher-order rewriting for \acs{lje}-rewriting, 
   \ie, rewriting in the body of iterators or quantifiers with respect tostrong 
   equality.This part of the theory is still under development. 
*}

constdefs
  (* cp0 is equivalent to cp \<Rightarrow> see thm cp0_eq_cp *)
  cp0    :: "([('\<tau>,'\<alpha>) VAL, '\<tau>] \<Rightarrow> '\<beta>) \<Rightarrow> bool"
  "cp0 P \<equiv> \<forall> Q \<tau>. P Q \<tau> = P (lift\<^sub>0 (Q \<tau>)) \<tau>"

  cp1    :: "((('\<tau>,'\<alpha>)VAL \<Rightarrow> ('\<tau>,'\<beta>)VAL) \<Rightarrow> ('\<tau>,'\<gamma>)VAL) \<Rightarrow> bool"
  "cp1 P \<equiv> \<forall> Q \<tau>. P Q \<tau> = P (\<lambda> x. (lift\<^sub>0 (Q x \<tau>))) \<tau>"

  cp2    :: "((('\<tau>,'\<alpha>)VAL \<Rightarrow> ('\<tau>,'\<beta>)VAL \<Rightarrow> ('\<tau>,'\<gamma>)VAL) \<Rightarrow> ('\<tau>,'\<delta>)VAL) \<Rightarrow> bool"
  "cp2 P \<equiv> \<forall> Q \<tau>. P Q \<tau> = P (\<lambda> x y. (lift\<^sub>0 (Q x y \<tau>))) \<tau>"


subsection {* Building a ruleset to unfold all cp-ness definitions *}

lemmas ss_cp_defs = cp_def cp0_def cp1_def cp2_def

text {* 
  The functional behaviour of @{text "lift\<^sub>0"}. This rule is sometimes
  useful in a proof.
*}

lemma lift0_apply: "lift\<^sub>0 X \<tau> = X"
  by(simp add: lift0_def)


subsection{* The rules to ``cp-unfold'' a term by hand *}

lemma cp0_unfold:
  "\<lbrakk> cp0 P; Q \<tau> = R \<tau> \<rbrakk> \<Longrightarrow> P Q \<tau> = P (lift\<^sub>0 (R \<tau>)) \<tau>"
  apply(unfold cp0_def, erule_tac x=Q in allE, erule_tac x=\<tau> in allE)
  by(simp add: lift0_apply)

lemma cp1_unfold:
  "\<lbrakk> cp1 P; \<And> x. Q x \<tau> = R x \<tau> \<rbrakk> \<Longrightarrow> P Q \<tau> = P (\<lambda> x. lift\<^sub>0 (R x \<tau>)) \<tau>"
  apply(unfold cp1_def, erule_tac x=Q in allE, erule_tac x=\<tau> in allE)
  by(simp add: lift0_apply)

lemma cp2_unfold:
  "\<lbrakk> cp2 P; \<And> x y. Q x y \<tau> = R x y \<tau> \<rbrakk> \<Longrightarrow> P Q \<tau> = P (\<lambda> x y. lift\<^sub>0 (R x y \<tau>)) \<tau>"
  apply(unfold cp2_def, erule_tac x=Q in allE, erule_tac x=\<tau> in allE)
  by(simp add: lift0_apply)

subsubsection{* The rules to ``cp-fold'' a term by hand *}

lemma cp0_fold:
  "\<lbrakk> cp0 P; Q \<tau> = R \<tau> \<rbrakk> \<Longrightarrow> P (lift\<^sub>0 (Q \<tau>)) \<tau>  = P R \<tau>"
  by(erule cp0_unfold[symmetric], simp)

lemma cp1_fold:
  "\<lbrakk> cp1 P; \<And> x. Q x \<tau> = R x \<tau> \<rbrakk> \<Longrightarrow> P (\<lambda> x. lift\<^sub>0 (Q x \<tau>)) \<tau>  = P R \<tau>"
  by(erule cp1_unfold[symmetric], simp)

lemma cp2_fold:
  "\<lbrakk> cp2 P; \<And> x y. Q x y \<tau> = R x y \<tau> \<rbrakk> \<Longrightarrow> P (\<lambda> x y. lift\<^sub>0 (Q x y \<tau>)) \<tau>  = P R \<tau>"
  by(erule cp2_unfold[symmetric], simp)


subsection{* Rules to Establish $\HolOclCp$ *}


text {* @{text "cp0"} rules *}

lemma cp0_const [simp]: "cp0 (\<lambda> x \<tau>. P)"
  by(simp add: cp0_def)

lemma cp0_id [simp]: "cp0 (\<lambda> x. x)"
  by(simp add: cp0_def lift0_apply)

lemma cp0_app:
  assumes cp0_1_P: "\<And> y. cp0 (\<lambda> x. (P x y))"
  and     cp0_2_P: "\<And> x. cp0 (P x)"
  shows            "cp0 (\<lambda> x. P x x)"
  apply(unfold cp0_def, clarify)
  apply(rule trans, rule_tac P="(\<lambda> x. P x ?Y)" and \<tau>=\<tau> in cp0_unfold)
  apply(rule cp0_1_P, rule refl)
  apply(rule trans, rule_tac P="(\<lambda> x. P ?X x)" and \<tau>=\<tau> in cp0_unfold)
  apply(rule cp0_2_P, rule refl)
  apply(rule refl)
  done

lemma cp0_of_cp0:
  assumes cp0_P: "cp0 P"
  and     cp0_Q: "cp0 Q"
  shows          "cp0 (\<lambda> x. P (Q x))"
  apply(unfold cp0_def, clarify)
  apply(rule trans, rule_tac P="(\<lambda> x. P x)" and \<tau>=\<tau> in cp0_unfold)
  apply(rule cp0_P, rule refl)
  apply(rule sym, rule trans, rule_tac P="(\<lambda> x. P x)" and \<tau>=\<tau> in cp0_unfold)
  apply(rule cp0_P)
  apply(rule_tac P="(\<lambda> x. Q x)" and \<tau>=\<tau> in cp0_fold)
  apply(rule cp0_Q, rule refl)
  apply(rule refl)
  done

lemma cp0_of_cp1:
  assumes cp1_P: "cp1 P"
  and   cp0_1_Q: "\<And> a. cp0 (\<lambda> x. Q x a)"
  shows          "cp0 (\<lambda> x. P (Q x))"
  apply(unfold cp0_def, clarify)
  apply(rule trans, rule_tac P="(\<lambda> x. P x)" and \<tau>=\<tau> in cp1_unfold)
  apply(rule cp1_P, rule refl)
  apply(rule sym, rule trans, rule_tac P="(\<lambda> x. P x)" and \<tau>=\<tau> in cp1_unfold)
  apply(rule cp1_P)
  apply(rule_tac P="(\<lambda> x. Q x ?X)" and \<tau>=\<tau> in cp0_fold)
  apply(rule cp0_1_Q, rule refl)
  apply(rule refl)
  done

lemma cp0_of_cp2:
  assumes cp2_P: "cp2 P"
  and   cp0_1_Q: "\<And> a b. cp0 (\<lambda> x. Q x a b)"
  shows          "cp0 (\<lambda> x. P (Q x))"
  apply(unfold cp0_def, clarify)
  apply(rule trans, rule_tac P="(\<lambda> x. P x)" and \<tau>=\<tau> in cp2_unfold)
  apply(rule cp2_P, rule refl)
  apply(rule sym, rule trans, rule_tac P="(\<lambda> x. P x)" and \<tau>=\<tau> in cp2_unfold)
  apply(rule cp2_P)
  apply(rule_tac P="(\<lambda> x. Q x ?X ?Y)" and \<tau>=\<tau> in cp0_fold)
  apply(rule cp0_1_Q, rule refl)
  apply(rule refl)
  done


text {* In the following, we prove the cp1 rules. *}

lemma cp1_const [simp]: "cp1 (\<lambda> x \<tau>. P)"
  by(simp add: cp1_def)

text {* 
  This rule corresponds to @{text "cp0_id"}. It has the same structure 
  except for the constant evaluation which is of course not needed
  for a nullary argument. 
*}
lemma cp1_const_eval [simp]: "cp1 (\<lambda> x. x X)"
  by(simp add: cp1_def lift0_apply)

lemma cp1_app:
  assumes cp1_1_P: "\<And> y. cp1 (\<lambda> x. (P x y))"
  and     cp1_2_P: "\<And> x. cp1 (P x)"
  shows            "cp1 (\<lambda> x. P x x)"
  apply(unfold cp1_def, clarify)
  apply(rule trans, rule_tac P="(\<lambda> x. P x ?Y)" and \<tau>=\<tau> in cp1_unfold)
  apply(rule cp1_1_P, rule refl)
  apply(rule trans, rule_tac P="(\<lambda> x. P ?X x)" and \<tau>=\<tau> in cp1_unfold)
  apply(rule cp1_2_P, rule refl)
  apply(rule refl)
  done

lemma cp1_of_cp0:
  assumes cp0_P: "cp0 P"
  and     cp1_Q: "cp1 Q"
  shows          "cp1 (\<lambda> x. P (Q x))"
  apply(unfold cp1_def, clarify)
  apply(rule trans, rule_tac P="(\<lambda> x. P x)" and \<tau>=\<tau> in cp0_unfold)
  apply(rule cp0_P, rule refl)
  apply(rule sym, rule trans, rule_tac P="(\<lambda> x. P x)" and \<tau>=\<tau> in cp0_unfold)
  apply(rule cp0_P)
  apply(rule_tac P="(\<lambda> x. Q x)" and \<tau>=\<tau> in cp1_fold)
  apply(rule cp1_Q, rule refl)
  apply(rule refl)
  done

lemma cp1_of_cp1:
  assumes cp1_P: "cp1 P"
  and   cp1_1_Q: "\<And> a. cp1 (\<lambda> x. Q x a)"
  shows          "cp1 (\<lambda> x. P (Q x))"
  apply(unfold cp1_def, clarify)
  apply(rule trans, rule_tac P="(\<lambda> x. P x)" and \<tau>=\<tau> in cp1_unfold)
  apply(rule cp1_P, rule refl)
  apply(rule sym, rule trans, rule_tac P="(\<lambda> x. P x)" and \<tau>=\<tau> in cp1_unfold)
  apply(rule cp1_P)
  apply(rule_tac P="(\<lambda> x. Q x ?X)" and \<tau>=\<tau> in cp1_fold)
  apply(rule cp1_1_Q, rule refl)
  apply(rule refl)
  done

lemma cp1_of_cp2:
  assumes cp2_P: "cp2 P"
  and   cp1_1_Q: "\<And> a b. cp1 (\<lambda> x. Q x a b)"
  shows          "cp1 (\<lambda> x. P (Q x))"
  apply(unfold cp1_def, clarify)
  apply(rule trans, rule_tac P="(\<lambda> x. P x)" and \<tau>=\<tau> in cp2_unfold)
  apply(rule cp2_P, rule refl)
  apply(rule sym, rule trans, rule_tac P="(\<lambda> x. P x)" and \<tau>=\<tau> in cp2_unfold)
  apply(rule cp2_P)
  apply(rule_tac P="(\<lambda> x. Q x ?X ?Y)" and \<tau>=\<tau> in cp1_fold)
  apply(rule cp1_1_Q, rule refl)
  apply(rule refl)
  done



text {* cp2 rules *}

lemma cp2_const [simp]: "cp2 (\<lambda> x \<tau>. P)"
  by(simp add: cp2_def)

lemma cp2_const_eval [simp]: "cp2 (\<lambda> x. x X Y)"
  by(simp add: cp2_def lift0_apply)

lemma cp2_app:
  assumes cp2_1_P: "\<And> y. cp2 (\<lambda> x. (P x y))"
  and     cp2_2_P: "\<And> x. cp2 (P x)"
  shows            "cp2 (\<lambda> x. P x x)"
  apply(unfold cp2_def, clarify)
  apply(rule trans, rule_tac P="(\<lambda> x. P x ?Y)" and \<tau>=\<tau> in cp2_unfold)
  apply(rule cp2_1_P, rule refl)
  apply(rule trans, rule_tac P="(\<lambda> x. P ?X x)" and \<tau>=\<tau> in cp2_unfold)
  apply(rule cp2_2_P, rule refl)
  apply(rule refl)
  done

lemma cp2_of_cp0:
  assumes cp0_P: "cp0 P"
  and     cp2_Q: "cp2 Q"
  shows          "cp2 (\<lambda> x. P (Q x))"
  apply(unfold cp2_def, clarify)
  apply(rule trans, rule_tac P="(\<lambda> x. P x)" and \<tau>=\<tau> in cp0_unfold)
  apply(rule cp0_P, rule refl)
  apply(rule sym, rule trans, rule_tac P="(\<lambda> x. P x)" and \<tau>=\<tau> in cp0_unfold)
  apply(rule cp0_P)
  apply(rule_tac P="(\<lambda> x. Q x)" and \<tau>=\<tau> in cp2_fold)
  apply(rule cp2_Q, rule refl)
  apply(rule refl)
  done

lemma cp2_of_cp1:
  assumes cp1_P: "cp1 P"
  and   cp2_1_Q: "\<And> a. cp2 (\<lambda> x. Q x a)"
  shows          "cp2 (\<lambda> x. P (Q x))"
  apply(unfold cp2_def, clarify)
  apply(rule trans, rule_tac P="(\<lambda> x. P x)" and \<tau>=\<tau> in cp1_unfold)
  apply(rule cp1_P, rule refl)
  apply(rule sym, rule trans, rule_tac P="(\<lambda> x. P x)" and \<tau>=\<tau> in cp1_unfold)
  apply(rule cp1_P)
  apply(rule_tac P="(\<lambda> x. Q x ?X)" and \<tau>=\<tau> in cp2_fold)
  apply(rule cp2_1_Q, rule refl)
  apply(rule refl)
  done

lemma cp2_of_cp2:
  assumes cp2_P: "cp2 P"
  and   cp2_1_Q: "\<And> a b. cp2 (\<lambda> x. Q x a b)"
  shows          "cp2 (\<lambda> x. P (Q x))"
  apply(unfold cp2_def, clarify)
  apply(rule trans, rule_tac P="(\<lambda> x. P x)" and \<tau>=\<tau> in cp2_unfold)
  apply(rule cp2_P, rule refl)
  apply(rule sym, rule trans, rule_tac P="(\<lambda> x. P x)" and \<tau>=\<tau> in cp2_unfold)
  apply(rule cp2_P)
  apply(rule_tac P="(\<lambda> x. Q x ?X ?Y)" and \<tau>=\<tau> in cp2_fold)
  apply(rule cp2_1_Q, rule refl)
  apply(rule refl)
  done

subsubsection {* Deciding cp using @{text "cp_unfold"} *}

text {*
   By converting a term into explicit $\HolOclCp$-representation one
  can also decide all context-passingness questions. Since the tactic
  uses a stepwise unfolding it can deal with cases where simp combined
  with chain rules for $\HolOclCp$-ness does not work. These are cases
  where currently the rule @{text "cp_compose2"} is applied by hand.
  The decision using @{text "cp_unfold"} works as follows:
  \begin{enumerate}
  \item apply @{text "cp_unfold"}
  \item if one of the rules @{text "cp0_cp_unfolded"}, @{text "cp1_cp_unfolded"},
        or @{text "cp2_cp_unfolded"} matches were done
        otherwise the $\HolOclCp$-ness proposition is false (assuming that
        the simpset of @{text "cp_unfold"} is complete with respect to the
        set of \OCL operations)
  \end{enumerate}
*}

lemma cp0_cp_unfolded:
  "cp0 (\<lambda> x \<tau>. P (lift\<^sub>0 (x \<tau>)) \<tau>)"
  by(simp add: cp0_def lift0_apply)

lemma cp1_cp_unfolded:
  "cp1 (\<lambda> Q \<tau>. P (\<lambda> x. lift\<^sub>0 (Q x \<tau>)) \<tau>)"
  by(simp add: cp1_def lift0_apply)

lemma cp2_cp_unfolded:
  "cp2 (\<lambda> Q \<tau>. P (\<lambda> x y. lift\<^sub>0 (Q x y \<tau>)) \<tau>)"
  by(simp add: cp2_def lift0_apply)



section {* Higher-Order lifting construction *}

constdefs
  lift1'           :: "[('\<gamma> \<Rightarrow> '\<eta> \<Rightarrow> '\<alpha>), '\<alpha> \<Rightarrow> '\<beta>] \<Rightarrow> ['\<gamma>, '\<eta>] \<Rightarrow> '\<beta>"
  "lift1' a1       \<equiv> (\<lambda> f x \<tau>. f (a1 x \<tau>))"

  lift2'           :: "[('\<alpha> \<Rightarrow> '\<gamma> \<Rightarrow> '\<eta>), ('\<beta> \<Rightarrow> '\<gamma> \<Rightarrow> 'f), ('\<eta> \<Rightarrow> 'f \<Rightarrow> '\<delta>)] \<Rightarrow> ['\<alpha>, '\<beta>, '\<gamma>] \<Rightarrow> '\<delta>"
  "lift2' a1 a2    \<equiv> (\<lambda> f x y \<tau>. f (a1 x \<tau>) (a2 y \<tau>))"

  lift3'           :: "['\<eta>\<Rightarrow>'h\<Rightarrow>'\<alpha>, 'f\<Rightarrow>'h\<Rightarrow>'\<beta>, 'g\<Rightarrow>'h\<Rightarrow>'\<gamma>, '\<alpha>\<Rightarrow>'\<beta>\<Rightarrow>'\<gamma>\<Rightarrow>'\<delta>] \<Rightarrow> ['\<eta>, 'f, 'g, 'h] \<Rightarrow> '\<delta>"
  "lift3' a1 a2 a3 \<equiv> (\<lambda> f x y z \<tau>. f (a1 x \<tau>) (a2 y \<tau>) (a3 z \<tau>))"

  lift_arg0        :: "('\<tau> \<Rightarrow> '\<beta>) \<Rightarrow> ('\<tau> \<Rightarrow> '\<beta>)"
  "lift_arg0       \<equiv> id"

  lift_arg1        :: "(('\<tau> \<Rightarrow> '\<gamma>) \<Rightarrow> ('\<tau> \<Rightarrow> '\<delta>)) \<Rightarrow> ('\<tau> \<Rightarrow> ('\<gamma> \<Rightarrow> '\<delta>))"
  "lift_arg1       \<equiv> (\<lambda> f \<tau> x. f (lift\<^sub>0 x) \<tau>)"

  lift_arg2        :: "(['\<alpha>\<Rightarrow>'\<beta>, '\<gamma>\<Rightarrow>'\<delta>] \<Rightarrow> ('\<eta>\<Rightarrow>'f)) \<Rightarrow> (['\<eta>, '\<beta>, '\<delta>] \<Rightarrow> 'f)"
  "lift_arg2       \<equiv> (\<lambda> f \<tau> x y. f (lift\<^sub>0 x) (lift\<^sub>0 y) \<tau>)"

subsection{* Rulesets to unfold the whole lifting constructions *}

lemmas ss_lifting  = lift1_def lift2_def lift3_def
                     lift1'_def lift2'_def lift3'_def
                     lift_arg0_def lift_arg1_def lift_arg2_def
                     DEF_def strictify_def isStrict_def
                     smash_def

lemmas ss_lifting' = lift0_def ss_lifting



subsection{* The Link to the old Combinator Set. *}
text {*
  Note, @{text"lift1"}, @{text "lift2"}, @{text "lift3"} are special
  cases of these general combinators.
*}
lemma lift1_by_lift1': "lift\<^sub>1 = (lift1' lift_arg0)"
  by(simp add: lift1_def lift1'_def lift_arg0_def)

lemma lift2_by_lift2': "lift\<^sub>2 = (lift2' lift_arg0 lift_arg0)"
  by(simp add: lift2_def lift2'_def lift_arg0_def)

lemma lift3_by_lift3': "lift3 = (lift3' lift_arg0 lift_arg0 lift_arg0)"
  by(simp add: lift3_def lift3'_def lift_arg0_def)



subsection{* Forward rules for the currently used constructions: *}

text {*
  As we have in the section about @{text "cp_unfold"}, transforming a term
  into explicit $\HolOclCp$-representation is very powerful. Its foundation
  is @{text "cp_unfold"} which relies on a simplifier set containing all
  context-passingess predicates for every argument of every \OCL operation.
  These set can be created by putting all forward rules from below whose first 
  assumption is unifyable with the definition of the \OCL function
  into this set.

  The @{text "fw_OCL_cp0_chain"} rules are the old @{text "OCL_cp"} rules
  who can be put into the standard simplifier set.
*}


section{* Context-passingness by construction *}
ML {*
local 
  val cpR_prover_simpset = simpset() addsimps ((thms "ss_lifting'") 
                                               @ (thms "ss_cp_defs"))
in
  fun cpR_prover mk_name (goal,arity,position,lifting) =
    let val (rname, name_prefix) = mk_name arity position lifting
        val thm  = prove_goalw (the_context()) [] goal
                   (fn prems => [ALLGOALS(simp_tac 
                                         (cpR_prover_simpset 
                                          addsimps prems))
                                ])
    in (store_thm(rname,thm), rname, name_prefix)
    end
end
*}

subsection{* pointwise context-passingness rules *}

ML {*
local
  fun mk_name_cpR arity position lifting =
      let val name_prefix = "OCL_cp" ^ (Int.toString arity) ^ "_"
                                     ^ (Int.toString position) ^ "_" 
      in  ("fw_" ^ name_prefix ^ lifting, name_prefix) end
in 
  val cpR_thms = map (cpR_prover mk_name_cpR) [
    ("f \<equiv> lift1 P \<Longrightarrow> cp0 f", 0, 1, "lift1"),
    ("f \<equiv> (lift1' lift_arg0) P \<Longrightarrow> cp0 f", 0, 1, "lift1'"),

    ("f \<equiv> lift2 P \<Longrightarrow> cp0 (\<lambda> x. f x Y)", 0, 1, "lift2"),
    ("f \<equiv> lift2 P \<Longrightarrow> cp0 (\<lambda> x. f X x)", 0, 2, "lift2"),
    ("f \<equiv> (lift2' lift_arg0 aY) P \<Longrightarrow> cp0 (\<lambda> x. f x Y)", 0, 1, "lift2'"),
    ("f \<equiv> (lift2' aX lift_arg0) P \<Longrightarrow> cp0 (\<lambda> x. f X x)", 0, 2, "lift2'"),
    ("f \<equiv> (lift2' aX lift_arg1) P \<Longrightarrow> cp1 (\<lambda> x. f X x)", 1, 2, "lift2'"), 
    ("f \<equiv> (lift2' aX lift_arg2) P \<Longrightarrow> cp2 (\<lambda> x. f X x)", 2, 2, "lift2'"),

    ("f \<equiv> lift3 P \<Longrightarrow> cp0 (\<lambda> x. f x Y Z)", 0, 1, "lift3"), 
    ("f \<equiv> lift3 P \<Longrightarrow> cp0 (\<lambda> x. f X x Z)", 0, 2, "lift3"), 
    ("f \<equiv> lift3 P \<Longrightarrow> cp0 (\<lambda> x. f X Y x)", 0, 3, "lift3"), 
    ("f \<equiv> (lift3' lift_arg0 aY aZ) P \<Longrightarrow> cp0 (\<lambda> x. f x Y Z)", 0, 1, "lift3'"),
    ("f \<equiv> (lift3' aX lift_arg0 aZ) P \<Longrightarrow> cp0 (\<lambda> x. f X x Z)", 0, 2, "lift3'"),
    ("f \<equiv> (lift3' aX aY lift_arg0) P \<Longrightarrow> cp0 (\<lambda> x. f X Y x)", 0, 3, "lift3'"),
    ("f \<equiv> (lift3' aX lift_arg1 aZ) P \<Longrightarrow> cp1 (\<lambda> x. f X x Z)", 1, 2, "lift3'"),
    ("f \<equiv> (lift3' aX lift_arg2 aZ) P \<Longrightarrow> cp2 (\<lambda> x. f X x Z)", 2, 2, "lift3'")
  ]
end
*}

subsection{* pointwise cp\_fold rules *}

ML {*
local
  fun mk_name_cpR_fold arity position lifting =
      let val name_prefix = "OCL_cp" ^ (Int.toString arity) ^ "_fold_" 
                                     ^ (Int.toString position) ^ "_" 
      in  ("fw_" ^ name_prefix ^ lifting, name_prefix) end
in 
  val cpR_fold_thms = map (cpR_prover mk_name_cpR_fold) [
    ("f \<equiv> lift1 P \<Longrightarrow> f (lift\<^sub>0 (X \<tau>)) \<tau> = f X \<tau>", 0, 1, "lift1"),
    ("f \<equiv> (lift1' lift_arg0) P \<Longrightarrow> f (lift\<^sub>0 (X \<tau>)) \<tau> = f X \<tau>", 0, 1, "lift1'"),

    ("f \<equiv> lift2 P \<Longrightarrow> f (lift\<^sub>0 (X \<tau>)) Y \<tau> = f X Y \<tau>", 0, 1, "lift2"),
    ("f \<equiv> lift2 P \<Longrightarrow> f X (lift\<^sub>0 (Y \<tau>)) \<tau> = f X Y \<tau>", 0, 2, "lift2"),
  
    ("f \<equiv> (lift2' lift_arg0 aY) P \<Longrightarrow> f (lift\<^sub>0 (X \<tau>)) Y \<tau> = f X Y \<tau>", 0, 1, "lift2'"),
    ("f \<equiv> (lift2' aX lift_arg0) P \<Longrightarrow> f X (lift\<^sub>0 (Y \<tau>)) \<tau> = f X Y \<tau>", 0, 2, "lift2'"),
    ("f \<equiv> (lift2' aX lift_arg1) P \<Longrightarrow> f X (\<lambda> x. lift\<^sub>0 (Y x \<tau>)) \<tau> = f X Y \<tau>", 1, 2, "lift2'"),
    ("f \<equiv> (lift2' aX lift_arg2) P \<Longrightarrow> f X (\<lambda> x y. lift\<^sub>0 (Y x y \<tau>)) \<tau> = f X Y \<tau>", 2, 2, "lift2'"),

    ("f \<equiv> lift3 P \<Longrightarrow> f (lift\<^sub>0 (X \<tau>)) Y Z \<tau> = f X Y Z \<tau>", 0, 1, "lift3"),
    ("f \<equiv> lift3 P \<Longrightarrow> f X (lift\<^sub>0 (Y \<tau>)) Z \<tau> = f X Y Z \<tau>", 0, 2, "lift3"),
    ("f \<equiv> lift3 P \<Longrightarrow> f X Y (lift\<^sub>0 (Z \<tau>)) \<tau> = f X Y Z \<tau>", 0, 3, "lift3"),

    ("f \<equiv> (lift3' lift_arg0 aY aZ) P \<Longrightarrow> f (lift\<^sub>0 (X \<tau>)) Y Z \<tau> = f X Y Z \<tau>", 0, 1, "lift3'"),
    ("f \<equiv> (lift3' aX lift_arg0 aZ) P \<Longrightarrow> f X (lift\<^sub>0 (Y \<tau>)) Z \<tau> = f X Y Z \<tau>", 0, 2, "lift3'"),
    ("f \<equiv> (lift3' aX aY lift_arg0) P \<Longrightarrow> f X Y (lift\<^sub>0 (Z \<tau>)) \<tau> = f X Y Z \<tau>", 0, 3, "lift3'"),
    ("f \<equiv> (lift3' aX lift_arg2 aZ) P \<Longrightarrow> f X (\<lambda> x y. lift\<^sub>0 (Y x y \<tau>)) Z \<tau> = f X Y Z \<tau>", 2, 2, "lift3'")
  ]
end

*}


subsection{* Chain rules for cp0 usable by simp or fast *}

text {* Unary \OCL operations: old lifting construction. *}

lemma fw_OCL_cp0_chain0_lift1:
  assumes f_def: "f \<equiv> lift1 P"
  and     cp0_X: "cp0 X"
  shows          "cp0 (\<lambda> x. f (X x))"
  by(insert cp0_X, simp add: f_def cp0_def lift1_def lift0_apply)

lemma fw_OCL_cp0_chain0_lift1':
  assumes f_def: "f \<equiv> (lift1' lift_arg0) P"
  and     cp0_X: "cp0 X"
  shows          "cp0 (\<lambda> x. f (X x))"
  by(insert cp0_X, simp add: f_def cp0_def lift_arg0_def lift1'_def lift0_apply)


text {* Binary \OCL operations: old lifting construction. *}

lemma fw_OCL_cp0_chain00_lift2:
  assumes f_def: "f \<equiv> lift2 P"
  and     cp0_X: "cp0 X"
  and     cp0_Y: "cp0 Y"
  shows          "cp0 (\<lambda> x. f (X x) (Y x))"
  by(insert cp0_X cp0_Y, 
     simp add: f_def cp0_def lift2_def lift0_apply)

lemma fw_OCL_cp0_chain00_lift2':
  assumes f_def: "f \<equiv> (lift2' lift_arg0 lift_arg0) P"
  and     cp0_X: "cp0 X"
  and     cp0_Y: "cp0 Y"
  shows          "cp0 (\<lambda> x. f (X x) (Y x))"
  by(insert cp0_X cp0_Y, 
     simp add: f_def cp0_def lift_arg0_def lift2'_def lift0_apply)

lemma fw_OCL_cp0_chain01_lift2':
  assumes f_def: "f \<equiv> (lift2' lift_arg0 lift_arg1) P"
  and     cp0_X: "cp0 X"
  and     cp0_Y: "\<And> a. cp0 (\<lambda> x. Y x a)"
  shows          "cp0 (\<lambda> x. f (X x) (Y x))"
  by(insert cp0_X cp0_Y, 
     simp add: f_def cp0_def lift_arg0_def lift_arg1_def lift2'_def lift0_apply)

lemma fw_OCL_cp0_chain02_lift2':
  assumes f_def: "f \<equiv> (lift2' lift_arg0 lift_arg2) P"
  and     cp0_X: "cp0 X"
  and     cp0_Y: "\<And> a b. cp0 (\<lambda> x. Y x a b)"
  shows          "cp0 (\<lambda> x. f (X x) (Y x))"
  by(insert cp0_X cp0_Y, 
     simp add: f_def cp0_def lift_arg0_def lift_arg2_def lift2'_def lift0_apply)


text {* Tertiary \OCL operations: $\isasymOclIterate$ and $\isasymMathOclIf$. *}

lemma fw_OCL_cp0_chain000_lift3:
  assumes f_def: "f \<equiv> lift3 P"
  and     cp0_X: "cp0 X"
  and     cp0_Y: "cp0 Y"
  and     cp0_Z: "cp0 Z"
  shows          "cp0 (\<lambda> x. f (X x) (Y x) (Z x))"
  by(insert cp0_X cp0_Y cp0_Z, 
     simp add: f_def cp0_def lift3_def lift0_apply)

lemma fw_OCL_cp0_chain000_lift3':
  assumes f_def: "f \<equiv> (lift3' lift_arg0 lift_arg0 lift_arg0) P"
  and     cp0_X: "cp0 X"
  and     cp0_Y: "cp0 Y"
  and     cp0_Z: "cp0 Z"
  shows          "cp0 (\<lambda> x. f (X x) (Y x) (Z x))"
  by(insert cp0_X cp0_Y cp0_Z, 
     simp add: f_def cp0_def lift_arg0_def lift3'_def lift0_apply)

lemma fw_OCL_cp0_chain020_lift3':
  assumes f_def: "f \<equiv> (lift3' lift_arg0 lift_arg2 lift_arg0) P"
  and     cp0_X: "cp0 X"
  and     cp0_Y: "\<And> a b. cp0 (\<lambda> x. Y x a b)"
  and     cp0_Z: "cp0 Z"
  shows          "cp0 (\<lambda> x. f (X x) (Y x) (Z x))"
  by(insert cp0_X cp0_Y cp0_Z, 
     simp add: f_def cp0_def lift_arg0_def lift_arg2_def lift3'_def)


subsection  {* 
  Chain rules for $\HolOclCp$ usable by \tactic{simp} or \tactic{fast}.
 *}

text {* Unary \OCL operations: old lifting construction. *}

lemma fw_OCL_cp_chain0_lift1':
  assumes f_def: "f \<equiv> (lift1' lift_arg0) P"
  and     cp_X: "cp X"
  shows          "cp (\<lambda> x. f (X x))"
  by(insert cp_X, simp add: f_def cp_by_cpify lift_arg0_def lift1'_def lift0_apply)


text {* Binary \OCL operations: old lifting construction. *}

lemma fw_OCL_cp_chain00_lift2':
  assumes f_def: "f \<equiv> (lift2' lift_arg0 lift_arg0) P"
  and     cp_X: "cp X"
  and     cp_Y: "cp Y"
  shows          "cp (\<lambda> x. f (X x) (Y x))"
  by(insert cp_X cp_Y, 
     simp add: f_def cp_by_cpify lift_arg0_def lift2'_def lift0_apply)

lemma fw_OCL_cp_chain01_lift2':
  assumes f_def: "f \<equiv> (lift2' lift_arg0 lift_arg1) P"
  and     cp_X: "cp X"
  and     cp_Y: "\<And> a. cp (\<lambda> x. Y x a)"
  shows          "cp (\<lambda> x. f (X x) (Y x))"
  by(insert cp_X cp_Y, 
     simp add: f_def cp_by_cpify lift_arg0_def lift_arg1_def lift2'_def lift0_apply)

lemma fw_OCL_cp_chain02_lift2':
  assumes f_def: "f \<equiv> (lift2' lift_arg0 lift_arg2) P"
  and     cp_X: "cp X"
  and     cp_Y: "\<And> a b. cp (\<lambda> x. Y x a b)"
  shows          "cp (\<lambda> x. f (X x) (Y x))"
  by(insert cp_X cp_Y, 
     simp add: f_def cp_by_cpify lift_arg0_def lift_arg2_def lift2'_def lift0_apply)


text {* Tertiary \OCL operations: $\isasymOclIterate$ and $\isasymMathOclIf$. *}

lemma fw_OCL_cp_chain000_lift3':
  assumes f_def: "f \<equiv> (lift3' lift_arg0 lift_arg0 lift_arg0) P"
  and     cp_X: "cp X"
  and     cp_Y: "cp Y"
  and     cp_Z: "cp Z"
  shows          "cp (\<lambda> x. f (X x) (Y x) (Z x))"
  by(insert cp_X cp_Y cp_Z, 
     simp add: f_def cp_by_cpify lift_arg0_def lift3'_def lift0_apply)

lemma fw_OCL_cp_chain020_lift3':
  assumes f_def: "f \<equiv> (lift3' lift_arg0 lift_arg2 lift_arg0) P"
  and     cp_X: "cp X"
  and     cp_Y: "\<And> a b. cp (\<lambda> x. Y x a b)"
  and     cp_Z: "cp Z"
  shows          "cp (\<lambda> x. f (X x) (Y x) (Z x))"
  by(insert cp_X cp_Y cp_Z, 
     simp add: f_def cp_by_cpify lift_arg0_def lift_arg2_def lift3'_def)


text {* Note, ll of these rules are registered for the use with
  \inlineholocl{ocl_setup_op} in the theory \theory{OCLL\_Kernel}.
*}

text {*
  @{text "cp0"} is equivalent to what we already have with @{text "cp"}.
  Thus it is currently added to the simpset to allow for
  a decision of @{text "cp0"}-questions based on the machinery already
  built for deciding @{text "cp"}-questions. This will change when the
  theories are cleaned up.
*}

lemma cp0_eq_cp:
  "cp0 P = cp P"
  by(simp add: cp0_def cp_by_cpify)


section {* Accessing Type Sums *}

constdefs
  FromL     :: "('\<alpha> + '\<beta>) up \<Rightarrow> '\<alpha> up"
  "FromL x  \<equiv> case x of lift v \<Rightarrow> ( sum_case (\<lambda> x. \<HolOclLiftLeft>x\<HolOclLiftRight>) (\<lambda> x. down) v)
  | down \<Rightarrow> down"

constdefs
  FromR      :: "('\<alpha> + '\<beta>) up \<Rightarrow> '\<beta> up"
  "FromR x  \<equiv> case x of lift v \<Rightarrow> ( sum_case (\<lambda>x. down ) (\<lambda> x. \<HolOclLiftLeft>x\<HolOclLiftRight>) v)
  | down \<Rightarrow> down"

lemma FromL_UU [simp]: "FromL \<bottom> = \<bottom>"
  by (simp add: FromL_def )

lemma FromR_UU [simp]: "FromR \<bottom> = \<bottom>"
  by (simp add: FromR_def)

lemma FromL_down [simp]: "FromL down = down"
  by (simp add: FromL_def)

lemma FromR_down [simp]: "FromR down = down"
  by (simp add: FromR_def)

lemma FromL_lift_Inl_id [simp]: "FromL \<HolOclLiftLeft>Inl x\<HolOclLiftRight> = \<HolOclLiftLeft>x\<HolOclLiftRight>"
  by (simp add: FromL_def)

lemma  FromR_lift_Inr_id [simp]: "FromR \<HolOclLiftLeft>Inr x\<HolOclLiftRight>  = lift x"
  by (simp add: FromR_def)

lemma FromR_Inl_UU [simp]: "FromR \<HolOclLiftLeft>Inl x\<HolOclLiftRight> = \<bottom>"
  by (simp add: FromR_def)

lemma FromL_Inr_UU [simp]: "FromL \<HolOclLiftLeft>Inr x\<HolOclLiftRight> = \<bottom>"
  by (simp add: FromL_def)


lemma h1: "DEF (FromL(x)) \<Longrightarrow> \<HolOclLiftLeft>Inl\<HolOclDropLeft>FromL x\<HolOclDropRight>\<HolOclLiftRight> = x "
  apply(simp add:  FromL_def lift_drop_idem) 
  apply(case_tac x, simp add: lift_drop_idem drop_lift_idem,auto)
  apply(case_tac \<alpha>, simp add: lift_drop_idem drop_lift_idem,auto)
  done

lemma h2: "DEF (FromL(x)) \<Longrightarrow> Inl\<HolOclDropLeft>FromL x\<HolOclDropRight> = \<HolOclDropLeft>x\<HolOclDropRight> "
  apply(simp add:  FromL_def lift_drop_idem) 
  apply(case_tac x, simp add: lift_drop_idem drop_lift_idem, auto)
  apply(case_tac \<alpha>, simp add: lift_drop_idem drop_lift_idem, auto)
  done


lemma DEF_FromL_exists_Inl: "DEF(FromL x) = (\<exists> y. x = \<HolOclLiftLeft>Inl y\<HolOclLiftRight>)"
  apply (simp add: FromL_def)
  apply (case_tac x, simp, auto)
  apply (case_tac \<alpha>, simp, auto)
  done


lemma DEF_FromR_exists_drop_Inr: "DEF (FromR(x)) \<Longrightarrow> \<HolOclLiftLeft>Inr \<HolOclDropLeft>FromR x\<HolOclDropRight>\<HolOclLiftRight> = x "
  apply(simp add:  FromR_def lift_drop_idem) 
  apply(case_tac x, simp add: lift_drop_idem drop_lift_idem, auto)
  apply(case_tac \<alpha>,simp add: lift_drop_idem drop_lift_idem, auto)
  done

lemma  FromL_Inl_inv [simp]: "(FromL \<HolOclLiftLeft>xb\<HolOclLiftRight> = \<HolOclLiftLeft>X\<HolOclLiftRight>) = (xb = Inl X)"
  apply (auto)
  apply (cases xb)
  apply (auto)
  done

lemma  FromR_Inr_inv [simp]: "(FromR \<HolOclLiftLeft>xb\<HolOclLiftRight> = \<HolOclLiftLeft>X\<HolOclLiftRight>) = (xb = Inr X)"
  apply (auto)
  apply (cases xb)
  apply (auto)
  done

lemma Inr_FromR_drop: "DEF (FromR(x)) \<Longrightarrow> Inr \<HolOclDropLeft>FromR x\<HolOclDropRight> = \<HolOclDropLeft>x\<HolOclDropRight>"
  apply(simp add:  FromR_def lift_drop_idem) 
  apply(case_tac x, simp add: lift_drop_idem drop_lift_idem,auto)
  apply(case_tac \<alpha>, simp add: lift_drop_idem drop_lift_idem,auto)
  done


lemma DEF_FromR_exits_Inr: "DEF (FromR x) = (\<exists>y. x = \<HolOclLiftLeft>Inr y\<HolOclLiftRight>)"
  apply (simp add: FromR_def)
  apply (case_tac x,simp,auto)
  apply (case_tac \<alpha>,simp,auto)
  done

lemma DEF_FromLD : "DEF(FromL x) \<Longrightarrow>  DEF(x)"
  apply (simp add: DEF_def FromL_def not_down_exists_lift)
  apply (case_tac x, simp, simp)
  done

lemma DEF_FromRD : "DEF(FromR x) \<Longrightarrow>  DEF(x)"
  apply (simp add: DEF_def FromR_def not_down_exists_lift)
  apply (case_tac x, simp, simp)
  done

end
