(*****************************************************************************
 *          HOL-OCL - a shallow embedding of OCL in Isabelle/HOL              
 *                                                                            
 * OCL_Kernel.thy - OCL Kernel
 * Copyright (C) 2001-2006 Achim D. Brucker <brucker@inf.ethz.ch>   
 *                         Burkhart Wolff   <bwolff@inf.ethz.ch>    
 *                                                                            
 * This file is part of HOL-OCL.                                              
 *                                                                            
 * HOL-OCL is free software; you can redistribute it and/or modify it under   
 * the terms of the GNU General Public License as published by the Free       
 * Software Foundation; either version 2 of the License, or (at your option)  
 * any later version.                                                         
 *                                                                            
 * HOL-OCL is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS 
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more 
 * details.                                                              
 *                                                                            
 * You should have received a copy of the GNU General Public License along    
 * with this program; if not, write to the Free Software Foundation, Inc.,    
 * 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.                  
 ******************************************************************************)

header {* The \OCL{} Kernel *}

theory OCL_Kernel 
imports 
  "$HOLOCL_HOME/src/Lifting" 
begin 
 text_raw{* \label{thy:OCL_Kernel} *}

section {* Global Syntax Infrastructure *}
text{* 
  In the theory files, this section contains the \acs{sml} code for 
  setting up the infra-strucure for switching between mathematical 
  and other notations for \OCL.
 *}
(*<*)
ML{* val OCL_strict_syntax_checking = ref(true);
     val TEST1 = ref([]:term list);
     val TEST2 = ref([]:term list);

     datatype OCL_syntax = std | math | ascii;
     fun OCL_syntax2string std   = "std"
        |OCL_syntax2string math  = "math"
        |OCL_syntax2string ascii = "ascii"
     val OCL_syntax_mode = ref ascii; 

     fun flip_args f (a::b::R) = f (b::a::R)

     fun trans_const mode name R =
         (if !OCL_strict_syntax_checking andalso !OCL_syntax_mode <> mode 
          then error(String.concat["strict checking: No ",
                                   OCL_syntax2string mode,
                                   " syntax allowed here"])
          else list_comb(Syntax.const name,R))   

     fun print_const name R = 
         let val name = ("_"^name^"_"^(OCL_syntax2string(!OCL_syntax_mode)))
         in  list_comb(Syntax.const name,R) end

     fun monop mode name (self::R) =
         (if !OCL_strict_syntax_checking andalso !OCL_syntax_mode <> mode 
          then error(String.concat["strict checking: No ",
                                   OCL_syntax2string mode,
                                   " syntax allowed here"])
          else list_comb(Syntax.const name,self::R))

     fun binop mode name (self::obj::R) =
         (if !OCL_strict_syntax_checking andalso !OCL_syntax_mode <> mode 
          then error(String.concat["strict checking: No ",
                                   OCL_syntax2string mode,
                                   " syntax allowed here"])
          else list_comb(Syntax.const name, self :: obj :: R))
 
     fun  emb f str = 
          let fun mk_name mode = "_"^str^"_"^(OCL_syntax2string mode)
          in  [(mk_name std,   f std str),
               (mk_name ascii, f ascii str),
               (mk_name math,  f math str)]
          end

     fun  emb_print str = (str, print_const str)

 *}

(*>*)
section {* ocl\_setup\_op *}
text {* 
  In the theory files, this section contains the \acs{sml} code that 
  implements the rulebase for the context-passingness rules that 
  are registered to the simplifier to decide $\HolOclCp$-ness
  in proofs \emph{without} the help of @{text "ocl_cp_unfold"}
   or @{text "ocl_simp"}.
*} 
(*<*)
ML{*

val OCL_setup_trace = ref true;

datatype setup_rule = RULE of
 {rule      : (string * thm),
  simp_tac  : clasimpset -> thm -> thm,
  is_atomic : thm -> bool, 
  register  : (string * thm) -> theory -> theory, 
  name_gen  : string -> string
};
  
fun OCL_setup_rule (name,thm) 
                   (RULE{rule,simp_tac,is_atomic,register,name_gen}) thy = 
    let val (rname,rthm) = rule
        fun warn s       = if !OCL_setup_trace 
                           then warning ("OCL_setup: "^s)
                           else ()
    in  let val thm'         = thm RS rthm
            val thm''        = simp_tac (claset_of thy, simpset_of thy) thm'
        in  writeln ((Context.theory_name thy)^ ".X." ^ name);
            if is_atomic thm'' 
            then register (name_gen name,thm'') thy
            else (warn ("resolvend of "^rname^" could not be simplified."); 
                 thy)
        end 
        handle _ => (warn ("rule "^rname^" does not match."); thy)
    end

*}

ML{*
fun default_namegen prefx name = 
       let val clean_dot = String.map (fn x => if x = #"." then #"_" else x)
       in  clean_dot(prefx^name) end


fun default_register (name,thm) thy = 
       fst(PureThy.add_thms [((name,thm),
                             [Simplifier.simp_add_global,
                              Classical.safe_intro_global])] thy) 

fun default_register_simponly (name,thm) thy = 
       (fst(PureThy.add_thms [((name,thm),
                             [Simplifier.simp_add_global])] thy)) 

fun default_register_claonly (name,thm) thy = 
       (fst(PureThy.add_thms [((name,thm),
                             [Classical.safe_intro_global])] thy)) 

fun register_to_simpset ss_ref (name,thm) thy =
    let val _ = (ss_ref := (!ss_ref) addsimps [thm])
    in  fst(PureThy.add_thms [((name,thm),[])] thy) end



val setup_rulebase = ref([]:setup_rule list);

fun add_setup_rules r = (setup_rulebase := r @ (!setup_rulebase))

*}

(*>*)

subsection {* Definition of the \acs{isar} interface to \protect\inlineocl|ocl_setup_op| *}
text {* 
 In the theory files, this section contains the \acs{sml} code that sets  
 up the isar syntax for command \inlineholocl+setup_op+.
*}
(*<*)
ML {*

structure isar_logic_setup =
struct 
structure P = OuterParse and K = OuterKeyword;

fun setup_opT defs thy =
    let fun KK X  = Library.apply(map (OCL_setup_rule X) (!setup_rulebase))
        fun KKK X = (X,get_thm thy (Name(X^"_def")))
    in  Library.apply(map (KK o KKK) defs) thy end


val setup_opP =
    OuterSyntax.command "ocl_setup_op" "setup for OCL operators"
                        OuterKeyword.thy_script
                        (P.$$$ "[" |-- (P.list1  (P.name) --| P.$$$ "]" ) >> (Toplevel.theory o setup_opT)
			);

val _ =  OuterSyntax.add_parsers [setup_opP];

end

 *}

(*>*)
subsection {* Setting up the Rulebase for $\HolOclCp$-reasoning. *}
text {*
 In the theory files, this section contains the \acs{sml} code that sets  
  up the rulebase for the context-passingness rules that are registered 
  to the simplifier to decide $\HolOclCp$-ness questions in interactive 
  proofs \emph{without} the help of @{text "ocl_cp_unfold"} or 
  @{text "ocl_simp"}.
*}
(*<*)
ML {*

val _ = let
  
  fun cp_rules name_prefix rname = RULE
    {rule      = (rname, thm rname),
     simp_tac  = fn _ => fn x => x,
     is_atomic = fn _ => true, 
     register  = default_register, 
     name_gen  = default_namegen name_prefix
    };

in

   add_setup_rules (
      map (cp_rules "OCL_cp_") [
          "cp_lift1_fw","cp_lift2_fw","cp_lift3_fw",
          "fw_OCL_cp_chain0_lift1'",
          "fw_OCL_cp_chain00_lift2'", "fw_OCL_cp_chain01_lift2'",
          "fw_OCL_cp_chain02_lift2'", 
          "fw_OCL_cp_chain000_lift3'", "fw_OCL_cp_chain020_lift3'"
      ]);

   add_setup_rules (
      map (cp_rules "OCL_cp0_chain_") [
          "fw_OCL_cp0_chain0_lift1", "fw_OCL_cp0_chain0_lift1'",
          "fw_OCL_cp0_chain00_lift2","fw_OCL_cp0_chain00_lift2'",
          "fw_OCL_cp0_chain01_lift2'","fw_OCL_cp0_chain02_lift2'",
          "fw_OCL_cp0_chain000_lift3", "fw_OCL_cp0_chain000_lift3'",
          "fw_OCL_cp0_chain020_lift3'"
      ])
end

*}
(*>*)
text{* 
  Registering the $\HolOclCp$-rules used to decide $\HolOclCp$-ness 
  questions in @{text "cp_unfold"} and the $\HolOclCp$-fold rules used 
  to convert a term in explicit $\HolOclCp$-representation to a term 
  in implicit $\HolOclCp$-representation.
*}
(*<*)
ML {*

val cpR_simpset      = ref(HOL_ss);
val cpR_fold_simpset = ref(HOL_ss);

val _ = 
    let fun cpR_rules ss_ref (thm, rname, name_prefix) = RULE
            {rule      = (rname, thm),
             simp_tac  = fn _ => fn x => x,
             is_atomic = fn _ => true, 
             register  = register_to_simpset ss_ref, 
             name_gen  = default_namegen name_prefix
            };
    in add_setup_rules (map (cpR_rules cpR_simpset) cpR_thms);
       add_setup_rules (map (cpR_rules cpR_fold_simpset) cpR_fold_thms)
    end

*}
(*>*)
end


