(* ****************************************************************************
 *          HOL-OCL - a shallow embedding of OCL in Isabelle/HOL              
 *                                                                            
 * OCL_universe.thy - Semantics of classes and class libraries
 * Copyright (C) 2001-2006 Achim D. Brucker <brucker@inf.ethz.ch>   
 *                         Burkhart Wolff   <bwolff@inf.ethz.ch>    
 *                                                                            
 * This file is part of HOL-OCL.                                              
 *                                                                            
 * HOL-OCL is free software; you can redistribute it and/or modify it under   
 * the terms of the GNU General Public License as published by the Free       
 * Software Foundation; either version 2 of the License, or (at your option)  
 * any later version.                                                         
 *                                                                            
 * HOL-OCL is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS 
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more 
 * details.                                                              
 *                                                                            
 * You should have received a copy of the GNU General Public License along    
 * with this program; if not, write to the Free Software Foundation, Inc.,    
 * 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.                  
 ***************************************************************************** *)

header {* The \OCL Universe *}
theory OCL_Universe 
imports 
  Datatype 
  "$HOLOCL_HOME/src/library/basic/OCL_Boolean_type" 
  "$HOLOCL_HOME/src/library/basic/OCL_String_type" 
  "$HOLOCL_HOME/src/library/basic/OCL_Real_type" 
  "$HOLOCL_HOME/src/library/basic/OCL_Integer_type" 
  "$HOLOCL_HOME/src/library/basic/OCL_Undefined"
  "$HOLOCL_HOME/src/library/OclAny/$OCLANY/OCL_OclAny_type" 
  "$HOLOCL_HOME/src/library/collection/$COLLECTION/OCL_Set_type" 
  "$HOLOCL_HOME/src/library/collection/$COLLECTION/OCL_Bag_type" 
  "$HOLOCL_HOME/src/library/collection/$COLLECTION/OCL_OrderedSet_type" 
  "$HOLOCL_HOME/src/library/collection/$COLLECTION/OCL_Sequence_type"  
begin
text_raw{* \label{thy:OCL_Universe} *}
hide const class 
text {* 
  Note, the definitions of \inlineocl{OclAny} differs,
  depending if \holocl is was built with a referential 
  universe or not.\footnote{This manual describes \holocl{} with  
    \universetype{} universes.}
*} 

section {* Foundations of the Universe Construction *}
text {*
  The \emph{universe} $\universe{U}$ is a family of \emphI{universe
  types} which comprise all value types\index{value types} (Real,
  Integer,String, Boolean, \ldots) and an extendible \emph{class type
  representation}\index{class type} induced by a class hierarchy.  The
  concepts of the universe construction has been presented in
  Chapt.~\ref{cha:faithf-repr-umlocl}.
*}

text {* 
  The ``initial'' universe type $\universe{U}_0$ is built as sum of
  the built-in value types and the type OclAny (including all its
  extensions representing subtypes of OclAny.
  *}

types 
  values   = "Real_0 + Integer_0 + Boolean_0 + String_0" 
  'a U     = "('a OclAny_0) + values"

types
	('a,'T) cl_ext = "('T \<times> 'a up) "
	('a,'b,'T) ext = "('a,'T)cl_ext + 'b"


text{* 
   The following setup of Isabelles syntax engine helps when
   presenting types corresponding to \OCL types:
*}

translations 
  "a U"       <= (type) "(OclAny_0 a) + (Real_0 + Integer_0 + Boolean_0 + String_0)"
  "Real_0"      <= (type) "real up"
  "Integer_0"   <= (type) "int up"
  "String_0"    <= (type) "char list up" (* string up *)
  "Boolean_0"   <= (type) "bool up" 



section {*Constants in and Operations on the Universe *}
text {* 
  The test operations for ``beeing a Real'' in a universe type were
  provided as follows: 
*}





constdefs
  sup        :: "(('a \<times> 'b )up) \<Rightarrow> 'a"
  "sup obj   \<equiv> fst((drop(obj)))"
  base       ::  "('a \<times> 'b up) up \<Rightarrow> 'b up"

 "base x  \<equiv> case x of lift (a,b)  => b
  | down => down"


constdefs
  refbase       ::  "('a \<times>  'b \<times> 'c up) up \<Rightarrow> 'c up"
 "refbase x  \<equiv> case x of lift (a,(b,c))  => c | down => down"

constdefs
  ref_of        ::  "('a \<times>  'b  \<times> 'c up) up \<Rightarrow> 'b"
 "ref_of x  \<equiv>         fst(snd(drop(x)))"


lemma DEF_baseD [simp]: "DEF(base(obj)) \<Longrightarrow> DEF(obj)"
  apply (simp add: DEF_def base_def not_down_exists_lift)
  apply (cases obj, simp)
  apply (auto)
  done
lemma DEF_baseD2 [simp]: "DEF(base(obj)) \<Longrightarrow> DEF(snd(drop(obj)))"
  apply (simp add: DEF_def base_def not_down_exists_lift)
  apply (cases obj, simp)
  apply (auto)
  done
lemma base_snd_drop: "DEF(base(obj)) \<Longrightarrow> base(obj) = snd(drop(obj))"
  apply (simp add: DEF_def base_def not_down_exists_lift)
  apply (cases obj, simp)
  apply (auto)
  done

lemma DEF_refbaseD [simp]: "DEF(refbase(obj)) \<Longrightarrow> DEF(obj)"
  apply (simp add: DEF_def refbase_def not_down_exists_lift)
  apply (cases obj, simp)
  apply (auto)
  done
lemma DEF_refbaseD2 [simp]: "DEF(refbase(obj)) \<Longrightarrow> DEF(snd(drop(obj)))"
  apply (simp add: DEF_def refbase_def not_down_exists_lift)
  apply (cases obj, simp)
  apply (auto)
  done
lemma refbase_snd_drop: "DEF(refbase(obj)) \<Longrightarrow> refbase(obj) = snd(snd(drop(obj)))"
  apply (simp add: DEF_def refbase_def not_down_exists_lift)
  apply (cases obj, simp)
  apply (auto)
  done



lemma base_UU [simp]: "base \<bottom> = \<bottom>"
  by (simp add: base_def )

lemma base_down [simp]: "base down = down"
  by (simp add: base_def)


lemma base_snd_lift [simp]: "base (lift(a,X)) = X"
  by (simp add: base_def)

lemma refbase_UU [simp]: "refbase \<bottom> = \<bottom>"
  by (simp add: refbase_def )

lemma refbase_down [simp]: "refbase down = down"
  by (simp add: refbase_def)


lemma refbase_snd_lift [simp]: "refbase (lift(a,(b,X))) = X"
  by (simp add: refbase_def)



types  'a OCLtype  = "'a U set"

     
constdefs  noext   :: "'a \<Rightarrow> ('a \<times> 'b up)"
          "noext x \<equiv> (x,\<bottom>)"


lemma fst_noext_id [simp]: "fst(noext x)=x"
  apply (simp (no_asm) add: noext_def)
  done
  
lemma snd_noext_UU [simp]: "snd(noext x)=\<bottom>"
  apply (simp (no_asm) add: noext_def)
done


end
